from __future__ import generator_stop

import os.path
from io import SEEK_END, SEEK_SET, BufferedIOBase, RawIOBase, TextIOBase, TextIOWrapper
from os import PathLike, fdopen, fspath, scandir
from sys import stdout
from typing import IO, TYPE_CHECKING, BinaryIO, Iterator, Optional, Tuple, Union, overload

from ._files import entrysuffix
from .iter import consume, iter_equal, resizer
from .math import PosInfInt
from .ops import logical_implication, logical_xor

if TYPE_CHECKING:
    from mmap import mmap
    from typing import Callable, Iterable, TextIO, TypeVar

    PathType = Union[str, PathLike[str]]
    Data = TypeVar("Data", str, bytes)

FILE_IO_BUFFER_SIZE = 8 * 1024 * 1024


class Empty(OSError):
    pass


def _check_arguments(mode, encoding=None):
    # type: (str, Optional[str]) -> Optional[str]

    is_text = "t" in mode
    is_binary = "b" in mode

    if not logical_xor(is_text, is_binary):
        raise ValueError(f"Explicit text or binary mode required: {mode}")

    if not logical_implication(is_binary, encoding is None):
        raise ValueError("Encoding is not None for binary file")

    if is_text:
        encoding = encoding or "utf-8"

    return encoding


def _stripmode(mode):
    return "".join(set(mode) - {"t", "b"})


def read_file(path, mode="b", encoding=None, errors=None):
    # type: (PathType, str, Optional[str], Optional[str]) -> bytes

    """Reads and returns whole file. If content is not needed use consume_file()"""

    encoding = _check_arguments(mode, encoding)

    if "r" not in mode:
        mode = "r" + mode

    with open(path, mode, encoding=encoding, errors=errors) as fr:
        return fr.read()


def write_file(data, path, mode="wb", encoding=None, errors=None):
    # type: (Union[str, bytes], PathType, str, Optional[str], Optional[str]) -> None

    """Writes file."""

    encoding = _check_arguments(mode, encoding)

    with open(path, mode, encoding=encoding, errors=errors) as fw:
        fw.write(data)


@overload
def read_or_raise(fin, size):
    # type: (IO[Data], int) -> Data

    ...


@overload
def read_or_raise(fin, size):
    # type: (mmap, int) -> bytes

    ...


def read_or_raise(fin, size):

    """Instead of returning a result smaller than `size` like `io.read()`,
    it raises and EOFError.
    """

    data = fin.read(size)
    if len(data) == 0:  # fixme: add `and size != 0`?
        raise Empty
    elif len(data) != size:
        raise EOFError

    return data


def get_file_range(path, start, size):
    # type: (PathType, int, int) -> bytes

    with open(path, "rb") as fp:
        fp.seek(start)
        return fp.read(size)


def truncate_file(path, size):
    # type: (PathType, int) -> None

    with open(path, "r+b") as fp:
        fp.truncate(size)


def wrap_text(bf, mode, encoding, errors, newline):
    if "t" in mode:
        tf = TextIOWrapper(bf, encoding, errors, newline)
        tf.mode = mode
        return tf
    return bf


def copen(file, mode="rt", archive_file=None, encoding=None, errors=None, newline=None, compresslevel=9, ext=None):
    # type: (Union[PathType, IO, int], str, Optional[str], Optional[str], Optional[str], Optional[str], int, Optional[str]) -> IO

    """Generic file open method. It supports transparent compression and improved text-mode handling.

    `file`: Can be a path, file-like or file descriptor.
    `mode`, `errors` and `newline`: see `io.open`
    `archive_file`: if a zip file is opened, this specifies the file within the archive.
    `encoding`: if `None` it defaults to "utf-8" in text-mode. It doesn't use any locale.
    `compresslevel`: 0-9, 0: no compression, 1: least, 9: highest compression.
            Only used if a compressed format is specified.
    `ext`: if `file` is a file-like object, than this can be one of {".gz", ".bz2", ".zip"}
            to enable transparent compression

    Returns a file-like object. If `file` was a fd,
            it will be closed when the file-like object is closed.
    """

    encoding = _check_arguments(mode, encoding)

    if isinstance(file, (str, PathLike)):
        ext = os.path.splitext(file)[1].lower()
    elif isinstance(file, int):
        file = fdopen(file, mode, encoding=encoding, errors=errors, newline=newline)
        ext = ext and ext.lower()
    else:
        ext = ext and ext.lower()

    if ext == ".gz":
        import gzip

        return gzip.open(file, mode, compresslevel=compresslevel, encoding=encoding, errors=errors, newline=newline)

    elif ext == ".bz2":
        import bz2

        return bz2.open(file, mode, compresslevel=compresslevel, encoding=encoding, errors=errors, newline=newline)

    elif ext == ".zip":
        if archive_file is None:
            raise ValueError("archive_file must be specified")

        from zipfile import ZipFile

        newmode = _stripmode(mode)

        with ZipFile(
            file, newmode
        ) as zf:  # note: even if the outer zip file is closed, the inner file can still be read apparently
            bf = zf.open(archive_file, newmode, force_zip64=True)

        return wrap_text(bf, mode, encoding, errors, newline)

    if isinstance(file, TextIOWrapper):
        return file

    if hasattr(file, "read") or hasattr(file, "write"):  # file should be in binary mode here
        return wrap_text(file, mode, encoding, errors, newline)

    return open(file, mode, encoding=encoding, errors=errors, newline=newline)


class OpenFileAndDeleteOnError:

    """Context manager which opens a file using the same arguments as `open`,
    but deletes the file in case an exception occurs after opening.
    """

    def __init__(self, file, mode="rt", encoding=None, errors=None, newline=None, compresslevel=9):
        # type: (PathType, str, Optional[str], Optional[str], Optional[str], int) -> None

        encoding = _check_arguments(mode, encoding)

        self.file = file
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.newline = newline
        self.compresslevel = compresslevel
        self.fp = None  # type: Optional[IO]

    def __enter__(self):
        # type: () -> IO

        self.fp = copen(self.file, self.mode, None, self.encoding, self.errors, self.newline, self.compresslevel)
        return self.fp

    def __exit__(self, exc_type, exc_value, traceback):
        self.fp.close()
        if exc_type:
            # fixme: race condition
            # use https://stackoverflow.com/a/3594593 and `ReOpenFile` on Windows
            # or SetFileInformationByHandle
            # see also: https://nullprogram.com/blog/2016/08/07/
            os.remove(self.file)


class OptionalWriteOnlyFile:
    def __init__(self, path=None, mode="xb", encoding=None, errors=None, newline=None, compresslevel=9):
        # type: (Optional[PathType], str, Optional[str], Optional[str], Optional[str], int) -> None

        if path:
            self.fp = copen(
                path, mode, encoding=encoding, errors=errors, newline=newline, compresslevel=compresslevel
            )  # type: Optional[IO]
        else:
            self.fp = None

    def __enter__(self):
        if self.fp:
            return self.fp
        else:
            return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.fp:
            self.fp.close()

    def write(self, data):
        pass

    def seek(self, offset, whence=None):
        pass


class StdoutFile:
    def __init__(self, path=None, mode="xb", encoding=None, errors=None, newline=None, compresslevel=9):
        # type: (Optional[PathType], str, Optional[str], Optional[str], Optional[str], int) -> None

        if path:
            self.fp = copen(path, mode, encoding=encoding, errors=errors, newline=newline, compresslevel=compresslevel)
            self.doclose = True
        else:
            self.doclose = False
            if "b" in mode:
                self.fp = stdout.buffer
            elif "t" in mode:
                self.fp = stdout
            else:
                raise ValueError(f"Explicit text or binary mode required: {mode}")

    def __enter__(self):
        return self.fp

    def __exit__(self, exc_type, exc_value, traceback):
        if self.doclose:
            self.fp.close()


class PathOrBinaryIO:
    def __init__(self, fname, mode="rb", close=False):
        # type: (Union[PathType, BinaryIO], str, bool) -> None

        if isinstance(fname, (RawIOBase, BufferedIOBase)):
            self.doclose = close  # type: bool
            self.fp = fname  # type: BinaryIO
        else:
            self.doclose = True
            self.fp = copen(fname, mode)

    def __enter__(self):
        # type: () -> IO

        return self.fp

    def __exit__(self, exc_type, exc_value, traceback):
        if self.doclose:
            self.fp.close()


class PathOrTextIO:
    def __init__(self, fname, mode="rt", encoding="utf-8", errors="strict", newline=None, close=False):
        # type: (Union[PathType, TextIO], str, str, str, Optional[str], bool) -> None

        if isinstance(fname, TextIOBase):
            self.doclose = close  # type: bool
            self.fp = fname  # type: TextIO
        else:
            self.doclose = True
            self.fp = copen(fname, mode, encoding=encoding, errors=errors, newline=newline)

    def __enter__(self):
        # type: () -> IO

        return self.fp

    def __exit__(self, exc_type, exc_value, traceback):
        if self.doclose:
            self.fp.close()


class LastLineFile:

    chunk_size = 1024 * 4
    nl = "\n"

    def __init__(self, path, mode="rt+"):
        self.f = open(path, mode)
        self.ll_pos = None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        self.f.close()

    def _seek_to_last_line(self):
        if self.ll_pos is None:
            self.ll_pos = self._get_last_line_pos()

        self.f.seek(self.ll_pos, SEEK_SET)

    def _get_last_line_pos(self):
        self.f.seek(0, SEEK_END)
        pos = self.f.tell()
        ret_pos = 0
        while pos > 0:
            seek_next = max(pos - self.chunk_size, 0)
            self.f.seek(seek_next, SEEK_SET)
            chunk = self.f.read(pos - seek_next)
            fpos = chunk.rfind("\n")
            if fpos != -1:
                ret_pos = seek_next + fpos + 1
                if ret_pos != pos:
                    break
            pos = seek_next
        return ret_pos

    def _validate(self, s):
        if self.nl in s:
            raise ValueError("Newline in input")

    def read(self):
        self._seek_to_last_line()
        return self.f.read()

    def newline(self, s):
        self.f.seek(0, SEEK_END)
        ret = self.f.write(self.nl)
        self.ll_pos = self.f.tell()
        return ret

    def replace(self, s):
        self._validate(s)
        self._seek_to_last_line()
        ret = self.f.write(s)
        self.f.truncate()
        return ret


class Tell:
    def __init__(self, fp):
        # type: (IO, ) -> None

        try:
            assert not fp.seekable()
        except AttributeError:
            pass

        self._fp = fp
        self._pos = 0

    # wrapped

    def read(
        self, size=None
    ):  # the docs for `io.BufferedIOBase` say that size=-1, but that's not true for `HTTPResponse`
        ret = self._fp.read(size)
        self._pos += len(ret)
        return ret

    def read1(self, n=-1):
        ret = self._fp.read1(n)
        self._pos += len(ret)
        return ret

    def readall(self):
        ret = self._fp.readall()
        self._pos += len(ret)
        return ret

    def readline(self, limit=-1):
        ret = self._fp.readline(limit)
        self._pos += len(ret)
        return ret

    def readlines(self, hint=-1):
        raise NotImplementedError
        return self._fp.readlines(hint)

    def readinto(self, b):
        ret = self._fp.readinto(b)
        self._pos += ret
        return ret

    def readinto1(self, b):
        ret = self._fp.readinto1(b)
        self._pos += ret
        return ret

    def write(self, b):  # fixme: should this be added to _pos?
        return self._fp.write(b)

    def writelines(self, lines):  # fixme: should this be added to _pos?
        return self._fp.writelines(lines)

    def seek(self, offset, whence=SEEK_SET):
        raise OSError("Stream is not seekable")

    def tell(self):
        return self._pos

    # redirected

    def close(self):
        return self._fp.close()

    def detach(self):
        return self._fp.detach()

    def fileno(self):
        return self._fp.fileno()

    def flush(self):
        return self._fp.flush()

    def isatty(self):
        return self._fp.isatty()

    def readable(self):
        return self._fp.readable()

    def seekable(self):
        return self._fp.seekable()

    def truncate(self, size=None):
        return self._fp.truncate(size)

    def writable(self):
        return self._fp.writable()


from collections import deque
from urllib import response


class BufferedTell(response.addinfourl):  # fixme: untested!!!
    def __init__(self, filesize):
        # response.addinfourl.__init__(self, *args, **kwargs)
        self._pos = 0
        self._read = 0
        self.filesize = filesize
        self.buf = deque(maxlen=1024)

    def read(self, num=-1):
        if num == -1:
            num = self.filesize - self._pos
            print("num=-1")
        if self._pos == self._read:
            self._pos += num
            self._read += num
            data = response.addinfourl.read(num)
            self.buf.extend(data)
            return data
        elif self._pos < self._read:
            delta = self._read - self._pos
            self._pos += num
            if num <= delta:
                data = b"".join(self.buf.pop() for i in range(delta))
                self.buf.extend(data[::-1])
                return data[-1 : -1 - num : -1]
            else:
                datab = b"".join(self.buf.pop() for i in range(delta))
                datab = datab[::-1]
                self.buf.extend(datab)
                dataf = response.addinfourl.read(num - delta)
                self.buf.extend(dataf)
                return datab + dataf
        else:
            # return b""
            raise Exception("Unbuffered read")

    def tell(self):
        return self._pos

    def seek(self, offset, whence=0):
        if whence == 0:
            self._pos = offset
        elif whence == 1:
            self._pos += offset
        elif whence == 2:
            self._pos = self.filesize - offset
        else:
            raise ValueError("Unknown whence")


def copyfilelike(fin, fout, amount=None, buffer=FILE_IO_BUFFER_SIZE, report=None):
    # type: (IO, IO, Optional[int], int, Optional[Callable]) -> int

    """Read data from `fin` in chunks of size `buffer` and write them to `fout`.
    Optionally limit the amount of data to `amount`. `report` can be a callable which receives
    the total number of bytes copied and bytes remaining.
    see `shutil.copyfileobj`
    """

    # todo: have different input and output buffers

    _amount = amount or PosInfInt

    copied = 0
    while _amount > 0:
        if report:
            report(copied, _amount + copied)

        data = fin.read(min(buffer, _amount))

        if not data:
            break
        fout.write(data)
        _amount -= len(data)
        copied += len(data)
    return copied


def simple_file_iter(fr, chunk_size=FILE_IO_BUFFER_SIZE):
    # type: (IO[Data], int) -> Iterator[Data]

    """Iterate file-like object `fr` and yield chunks of size `chunk_size`."""

    assert isinstance(chunk_size, int), "chunk_size needs to be an integer"

    # return iter(partial(fr.read, chunk_size), "") sentinel cannot be defined well
    while True:
        data = fr.read(chunk_size)
        if data:
            yield data
        else:
            break


def reversed_file_iter(fp, chunk_size=FILE_IO_BUFFER_SIZE):
    # type: (IO[Data], int) -> Iterator[Data]

    """Generate blocks of file's contents in reverse order."""

    fp.seek(0, SEEK_END)
    here = fp.tell()
    while here > 0:
        delta = min(chunk_size, here)
        here -= delta
        fp.seek(here, SEEK_SET)
        yield fp.read(delta)


def limited_file_iter(fr, amount, chunk_size=FILE_IO_BUFFER_SIZE):
    # type: (IO[Data], int, int) -> Iterator[Data]

    """Iterate file-like object `fr` and yield chunks of size `chunk_size`.
    Limit output to `amount` bytes.
    """

    while amount > 0:
        data = fr.read(min(chunk_size, amount))
        if not data:
            break
        yield data
        amount -= len(data)


def iterfilelike(fr, start=0, amount=None, chunk_size=FILE_IO_BUFFER_SIZE):
    # type: (IO[Data], int, Optional[int], int) -> Iterable[Data]

    """Iterate file-like object `fr` and yield chunks of size `chunk_size`.
    Starts reading at `start` and optionally limit output to `amount` bytes.
    """

    if start:
        fr.seek(start)

    if amount is None:
        return simple_file_iter(fr, chunk_size)
    else:
        return limited_file_iter(fr, amount, chunk_size)


def blockfileiter(path, mode="rb", encoding=None, errors=None, start=0, amount=None, chunk_size=FILE_IO_BUFFER_SIZE):
    # type: (PathType, str, Optional[str], Optional[str], int, Optional[int], int) -> Iterator[Union[str, bytes]]

    """Iterate over file at `path` and yield chunks of size `chunk_size`.
    Optionally limit output to `amount` bytes.
    """

    encoding = _check_arguments(mode, encoding)

    with open(path, mode, encoding=encoding, errors=errors) as fr:
        yield from iterfilelike(fr, start, amount, chunk_size)


def blockfilesiter(paths, chunk_size=FILE_IO_BUFFER_SIZE):
    # type: (Iterable[PathType], int) -> Iterator[bytes]

    """Iterate over a list of files and return their contents in a
    continuous stream of chunks of size `chunk_size`.
    That means one chunk can span multiple files.
    """

    chunk = b""

    for path in paths:
        with open(path, "rb") as fr:
            if len(chunk) != 0:
                # print("a")
                data = fr.read(chunk_size - len(chunk))
                chunk += data  # fixme: probably slow for lots of small files

            if len(chunk) == chunk_size:
                # print("b")
                yield chunk
                chunk = b""

            if len(chunk) == 0:
                # print("c")
                for data in iterfilelike(fr, chunk_size=chunk_size):  # is chunk overwritten here?
                    if len(data) == chunk_size:
                        yield data
                    else:
                        chunk = data

    if chunk:
        yield chunk


def bufferedfileiter(
    path, chunk_size, mode="rb", encoding=None, errors=None, start=0, amount=None, buffer_size=FILE_IO_BUFFER_SIZE
):
    # type: (PathType, int, str, Optional[str], Optional[str], int, Optional[int], int) -> Iterable[Union[str, bytes]]

    """Iterate over file at `path` reading chunks of size `buffer_size` at a time.
    Yields data chunks of `chunk_size` and optionally limits output to `amount` bytes.
    """

    return resizer(blockfileiter(path, mode, encoding, errors, start, amount, buffer_size), chunk_size)


def byte_out(path, buffer_size=FILE_IO_BUFFER_SIZE):
    # type: (PathType, int) -> Iterable[bytes]

    """Reads file at `path` byte by byte, using a read buffer of size `buffer_size`."""

    return resizer(blockfileiter(path, mode="rb", chunk_size=buffer_size), 1)


def consume_file(filename, buffer_size=FILE_IO_BUFFER_SIZE):
    # type: (PathType, int) -> None
    """reads whole file but ignores content"""

    consume(blockfileiter(filename, mode="rb", chunk_size=buffer_size))


# was: same_files, textfile_equal: equal_files(*paths, mode="rt")
def equal_files(*paths, mode="rb", encoding=None, errors=None, amount=None, chunk_size=FILE_IO_BUFFER_SIZE):
    # type: (*PathType, str, Optional[str], Optional[str], Optional[int], int) -> bool

    """Check if files at `*paths` are equal. Chunks of size `chunk_size` are read at a time.
    Data can be optionally limited to `amount`.
    """

    its = tuple(
        blockfileiter(path, mode=mode, encoding=encoding, errors=errors, amount=amount, chunk_size=chunk_size)
        for path in paths
    )
    return iter_equal(*its)


def is_all_byte(fr, thebyte=b"\0", chunk_size=FILE_IO_BUFFER_SIZE):
    # type: (IO, bytes, int) -> bool

    """Test if file-like `fr` consists only of `thebyte` bytes."""

    assert isinstance(thebyte, bytes)

    thebyte = thebyte * chunk_size
    for data in simple_file_iter(fr, chunk_size):
        if data != thebyte[: len(data)]:
            return False
    return True


def iter_zip(file, mode="rb", encoding=None, errors=None, newline=None, password=None):
    # type: (Union[PathType, IO], str, Optional[str], Optional[str], Optional[str], Optional[bytes]) -> Iterator[Tuple[str, IO]]

    """
    Iterate file-pointers to archived files. They are valid for one iteration step each.
    If `file` is a file-like, it must be seekable.
    """

    # from pyzipper import AESZipFile as ZipFile
    from zipfile import ZipFile

    encoding = _check_arguments(mode, encoding)
    newmode = _stripmode(mode)

    with ZipFile(file, newmode) as zf:
        for zi in zf.infolist():
            if not zi.is_dir():
                with zf.open(zi, newmode, password) as bf:
                    yield zi.filename, wrap_text(bf, mode, encoding, errors, newline)


def iter_7zip(
    file: Union[BinaryIO, str, PathLike],
    mode: str = "rb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    password: Optional[str] = None,
) -> Iterator[Tuple[str, IO]]:

    from py7zr import SevenZipFile

    encoding = _check_arguments(mode, encoding)
    newmode = _stripmode(mode)

    with SevenZipFile(file, mode=newmode) as zf:
        for fname, bf in zf.readall().items():
            yield fname, wrap_text(bf, mode, encoding, errors, newline)


def iter_tar(file, mode="rb", encoding=None, errors=None, newline=None):
    # type: (Union[PathType, BinaryIO], str, Optional[str], Optional[str], Optional[str]) -> Iterator[Tuple[str, IO]]

    """
    Iterate file-pointers to archived files. They are valid for one iteration step each.
    With the correct mode, `file` can be a non-seekable file-like.
    """

    import tarfile

    encoding = _check_arguments(mode, encoding)
    newmode = _stripmode(mode)

    if isinstance(file, PathLike):
        file = fspath(file)

    if isinstance(file, str):
        cls = tarfile.open(name=file, mode=newmode + "|*")
    else:
        cls = tarfile.open(fileobj=file, mode=newmode + "|*")

    with cls as tf:
        for ti in tf:
            if ti.isfile():  # fixme: what about links?
                bf = tf.extractfile(ti)
                assert bf, "member `ti` is file, but still `extractfile` returned `None`"
                with bf:  # no `mode` for `extractfile`
                    yield ti.name, wrap_text(bf, mode, encoding, errors, newline)


def iter_dir(
    path: str,
    mode: str = "rb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    follow_symlinks: bool = True,
    archives: bool = False,
    joiner: str = "/",
) -> Iterator[Tuple[str, IO]]:

    """
    Iterate file-pointers to files in directory `path`.
    They are valid for one iteration step each.
    If `archives` is True, archives like zip files will be traversed as well.
    """

    encoding = _check_arguments(mode, encoding)

    archive_funcs = {
        ".zip": iter_zip,
        ".tar": iter_tar,
        ".tgz": iter_tar,  # what about tar.gz
        ".tbz": iter_tar,  # what about tar.bz2
        ".7z": iter_7zip,
    }

    if archives:
        open = copen

    with scandir(path) as scan:
        for entry in scan:
            if entry.is_file(follow_symlinks=follow_symlinks):

                ext = entrysuffix(entry).lower()
                iter_archive = archive_funcs.get(ext, None)
                if archives and iter_archive:
                    for name, fr in iter_archive(entry.path, mode, encoding, errors, newline):
                        yield entry.path + joiner + name, fr
                else:
                    with open(entry.path, mode, encoding=encoding, errors=errors, newline=newline) as fr:
                        yield entry.path, fr

            elif entry.is_dir(follow_symlinks=follow_symlinks):
                yield from iter_dir(entry.path, mode, encoding, errors, newline, follow_symlinks, archives)


def iter_lines(path, encoding="utf-8", errors="strict", newline=None, verbose=False):
    # type: (str, str, str, Optional[str], bool) -> Iterator[str]

    """Yield lines from text file. Handles compressed files as well.
    if verbose is True, print a progress bar
    """

    with copen(path, "rt", encoding=encoding, errors=errors, newline=newline) as fr:

        if verbose:
            from tqdm import tqdm

            prev_pos = 0

            fr.seek(0, os.SEEK_END)
            total = fr.tell()
            fr.seek(0, os.SEEK_SET)

            with tqdm(total=total, leave=False, unit="B", unit_scale=True) as pbar:
                while True:  # cannot use `for line in fr` here
                    line = fr.readline()
                    if not line:
                        break
                    cur_pos = fr.tell()
                    pbar.update(cur_pos - prev_pos)
                    yield line
                    prev_pos = cur_pos
        else:
            for line in fr:
                yield line


def iter_stripped(path, encoding="utf-8", errors="strict", newline=None):
    # type: (str, str, str, Optional[str]) -> Iterator[str]

    """Yield lines from text file with trailing newlines stripped.
    Handles compressed files as well.
    """

    with copen(path, "rt", encoding=encoding, errors=errors, newline=newline) as fr:
        for line in fr:
            yield line.rstrip("\r\n")


# is this still needed?
def file_byte_reader(filename, inputblocksize, outputblocksize, DEBUG=True):
    # type: (PathType, int, int, bool) -> Iterator[bytes]

    assert (inputblocksize % outputblocksize == 0) or (
        outputblocksize % inputblocksize == 0
    ), "Neither input nor output size is a multiple of the other"

    bytes_yielded = 0
    bytes = bytearray(max(inputblocksize, outputblocksize))
    bytes_used = 0
    for read in blockfileiter(filename, chunk_size=inputblocksize):
        bytes[bytes_used : bytes_used + len(read)] = read
        # print("read {} bytes to pos [{}:{}]".format(len(read),bytes_used,bytes_used+len(read)))
        bytes_used += len(read)
        pos = 0
        while bytes_used >= outputblocksize:
            yield bytes[pos : pos + outputblocksize]
            bytes_yielded += outputblocksize
            # print("yielded {} bytes from pos [{}:{}]".format(outputblocksize,pos,pos+outputblocksize))
            pos += outputblocksize
            bytes_used -= outputblocksize
    if DEBUG:
        import os.path

        filesize = os.path.getsize(filename)
        if filesize != bytes_yielded:
            print(f"{bytes_yielded} bytes yielded, filesize: {filesize}")



from __future__ import generator_stop

import logging
import random
import sys
from collections import deque
from functools import partial
from itertools import chain, combinations, count, groupby, islice, product, repeat, starmap, tee, zip_longest
from operator import add, itemgetter
from random import randrange
from time import sleep, time
from types import GeneratorType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Container,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    TextIO,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from .exceptions import EmptyIterable, IteratorExhausted
from .ops import operator_in

if TYPE_CHECKING:
    from .typing import Orderable, SizedIterable

    OrderableT = TypeVar("OrderableT", bound=Orderable)
    ExceptionsType = Union[Type[Exception], Tuple[Type[Exception], ...]]

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")

logger = logging.getLogger(__name__)


def one(it: Iterator[T]) -> T:

    try:
        ret = next(it)
    except StopIteration:
        raise EmptyIterable()
    try:
        next(it)
    except StopIteration:
        pass
    else:
        raise ValueError("Iterator yielded more than one value")

    return ret


def iterrandrange(a: int, b: int) -> Iterator[int]:

    """Yields a stream of non-cryptographic random numbers between `a` (inclusive) and `b` (exclusive)."""

    while True:
        yield randrange(a, b)  # nosec


def repeatfunc(func, times=None, *args):
    # type: (Callable[..., T], Optional[int], *Any) -> Iterator[T]

    """Repeat calls to func with specified arguments.
    Example:  repeatfunc(random.random)
    See: https://docs.python.org/3/library/itertools.html#itertools-recipes
    """

    if times is None:
        return starmap(func, repeat(args))
    return starmap(func, repeat(args, times))


def _lstr(it: Iterable, length: Optional[int] = None) -> Tuple[Optional[int], str]:

    if length is None:
        try:
            length = len(it)  # type: ignore[arg-type]
        except TypeError:
            pass

    if length is not None:
        lstr = f"/{length}"
    else:
        lstr = ""

    return length, lstr


def progress(it, length=None, refresh=1, end="\r", file=sys.stdout, extra_info_callback=None, disable=None, delta=1):
    # type: (Union[Iterable, Collection], Optional[int], float, str, TextIO, Optional[Callable[[int, Optional[int]], str]], Optional[bool], Optional[int]) -> Iterator

    """Wraps an iterable `it` to periodically print the progress every `refresh` seconds.
    `lengths` is the total size of `it`. `len(it)` is used to get the size if set to `None`.
    `refresh` updates the printed output every `refresh` seconds.
    `delta` specifies the value which is added to the counter every iteration.
            `None` uses `len(elm)` to determine the value. Defaults to 1.
    """

    """ todo: `from operator import length_hint`
        Use `length_hint(it)` for a guess which might be better than nothing.
        Can return 0.
    """

    if delta is not None and delta < 1:
        raise ValueError(f"`delta` must be a integer larger than 1 or `None`, not '{delta}'")

    if disable is None and hasattr(file, "isatty") and not file.isatty():
        disable = True

    if disable:
        yield from it
        return

    length, lstr = _lstr(it, length)
    extra = ""
    last = start = time()
    total = 0

    try:
        for elm in it:
            yield elm
            if delta is None:
                total += len(elm)
            else:
                total += delta
            current = time()
            if current - last > refresh:
                last = current
                duration = current - start
                if extra_info_callback:
                    extra = extra_info_callback(total, length)
                    extra = f" [{extra}]"
                print(
                    f"{total}{lstr}, running for {int(duration)} seconds ({total/duration:0.2e}/s){extra}.",
                    end="\r",
                    file=file,
                )
    except KeyboardInterrupt:
        duration = last - start
        if duration > 0:
            print(
                f"Unsafely aborted after reading {total}{lstr} in {int(duration)} seconds ({total/duration:0.2e}/s).",
                end=end,
                file=file,
            )
        else:
            print(f"Unsafely aborted after reading {total}{lstr} in {int(duration)} seconds.", end=end, file=file)
        raise
    else:
        duration = last - start
        if duration > 0:
            print(f"Finished {total} in {int(duration)} seconds ({total/duration:0.2e}/s).", end=end, file=file)
        else:
            print(f"Finished {total} in {int(duration)} seconds.", end=end, file=file)


def progressdata(it, length=None, refresh=1, end="\r", file=sys.stdout, extra_info_callback=None, disable=None):
    # type: (Union[Iterable, Collection], Optional[int], float, str, Optional[TextIO], Optional[Callable[[int, Optional[int]], str]], Optional[bool]) -> Iterable

    return progress(it, length, refresh, end, file, extra_info_callback, disable, None)


class Progress:

    """Cannot create a decorator to do this to generic generators,
    because attributes cannot be set on generator objects.
    """

    def __init__(self, obj, refresh=1):
        # type: (Collection, float) -> None

        self.obj = obj
        self.refresh = refresh

    def __iter__(self):
        # type: () -> Iterator

        return progress(self.obj, len(self.obj), self.refresh)

    def __len__(self):
        # type: () -> int

        """Can raise a TypeError if underlying object does not support `len()`."""

        return len(self.obj)


def count_distinct(it):
    # type: (Iterable[Any], ) -> int

    return len(set(it))


def reversedzip(*its):
    return zip(*map(reversed, its))


def flatten(it):
    # type: (Union[tuple, list], ) -> Iterator

    for elm in it:
        if isinstance(elm, (tuple, list)):
            yield from flatten(elm)
        else:
            yield elm


def diffsgn(a, b):
    # type: (OrderableT, OrderableT) -> int

    """Sign of the finite derivative.
    Equal to `-cmp(a, b)` or `cmp(b, a).
    """

    return (a < b) - (a > b)


def asc_peaks(iterable):
    # type: (Iterable[OrderableT], ) -> Iterator[OrderableT]

    """Yields ascending peaks from `iterable`."""

    it = iter(iterable)
    try:
        cur = next(it)
    except StopIteration:
        return

    for x in it:
        if x < cur:
            yield cur
        cur = x
    yield cur


def extrema(iterable, first={1, -1}, last={1, -1}, derivatives={1, -1}):
    # type: (Iterable[OrderableT], Set[int], Set[int], Set[int]) -> Iterator[OrderableT]

    it = iter(iterable)
    try:
        old = next(it)
    except StopIteration:
        return
    try:
        old_2 = next(it)
        old_d = diffsgn(old, old_2)
    except StopIteration:
        yield old  # fixme: should this be returned?
        return

    if old_d in first:
        yield old

    old = old_2

    for new in it:
        new_d = diffsgn(old, new)
        if new_d != old_d and old_d in derivatives:
            yield old
        old = new
        old_d = new_d

    if old_d in last:
        yield old


def peaks(it):
    # type: (Iterable[OrderableT], ) -> Iterator[OrderableT]

    """Yields peaks of `it`."""

    return extrema(it, {-1}, {1}, {1})


def valleys(it):
    # type: (Iterable[OrderableT], ) -> Iterator[OrderableT]

    """Yields valleys of `it`."""

    return extrema(it, {1}, {-1}, {-1})


def empty():
    # type: () -> Iterator

    """An empty iterator. Other methods include:
    iter(())
    yield from ()
    """

    return
    yield  # type: ignore # pylint: disable=unreachable


def lastdefault(it, default=None):
    # type: (Iterable[T], Optional[T]) -> Optional[T]

    """Returns the last element of iterable `it`. It discards all other values.
    This method is about 1/2 times as fast as `last()` for large iterables.
    """

    ret = default
    for i in it:
        ret = i
    return ret


def last(it):
    # type: (Iterable[T], ) -> T

    """Returns the last element of iterable `it`. It discards all other values.
    This method is about 2 times as fast as `lastdefault()` for large iterables.
    """

    try:
        return deque(it, 1).pop()
    except IndexError:
        raise EmptyIterable("Empty iterable") from None


def batch(it, n, func=None):
    # type: (Iterable[Any], int, Optional[Callable]) -> Iterator[Any] # return type cannot be more specific because of filter()

    """Batches iterable `it` into batches of size `n`.
    Optionally post-processes batch with `func`.
    """

    it = iter(it)
    while True:
        chunk_it = islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return

        batch_iter = chain((first_el,), chunk_it)

        if func:
            batch_iter = func(batch_iter)

        yield batch_iter


def advance(it, n):
    # type: (Iterable[T], int) -> None

    """Advances the iterable `it` n steps."""

    """
    for i in range(n):
        next(it, None)
    """
    next(islice(it, n, n), None)


def filternone(it):
    # type: (Iterable, ) -> Iterator

    """Removes all `None` values from `it`."""

    return (i for i in it if i is not None)


def all_equal(it):
    # type: (Iterable, ) -> bool

    """Returns `True` if all elements of `it` are equal to each other.

    Much faster than:
            return len(set(it)) in (0, 1)
    and also faster than:
            try:
                    first = next(it)
            except StopIteration:
                    return True
            return all(first == rest for rest in it)
    for both extreme cases: all equal and none equal
    """

    g = groupby(it)
    return next(g, True) and not next(g, False)  # type: ignore


def pairwise(it):
    # type: (Iterable[T], ) -> Iterator[Tuple[T, T]]

    """Return two elements of iterable `it` at a time,
    but only advances the iterable by 1 each time.
    pairwise((a, b, c)) -> (a, b), (b, c)
    """

    a, b = tee(it, 2)
    next(b, None)
    return zip(a, b)


def x_wise(it, size):
    # type: (Iterable, int) -> Iterator

    """Returns `size` elements of iterable `it`, but only advances the iterable by 1 each time.
    ([1,2,3,4,5], 2) -> [1,2], [2,3], [3,4], [4,5]
    ([1,2,3,4,5], 3) -> [1,2,3], [2,3,4], [3,4,5]
    ([1,2,3,4,5], 4) -> [1,2,3,4], [2,3,4,5]
    and so only
    """

    copies = tee(it, size)
    for copy, pos in zip(copies, range(size)):
        advance(copy, pos)
    return zip(*copies)


triples = partial(x_wise, size=3)


def findfirst(func, it, default=(None, None)):
    # type: (Callable[[T], bool], Iterable[T], Tuple[Optional[int], Optional[T]]) -> Tuple[Optional[int], Optional[T]]

    """Find the first element of iterable `it` where `func(elm)` evaluates to True.
    Return `default` if not such element was found.
    """

    for i, x in enumerate(it):
        if func(x):
            return i, x

    return default


def is_empty(it):
    # type: (Iterator, ) -> bool

    """Returns True if the iterator `it` is already fully consumed.
    If not, it will be advanced.
    """

    try:
        next(it)
        return False
    except StopIteration:
        return True


def consume(it):
    # type: (Iterable, ) -> None

    """Consumes the iterable `it` completely."""

    deque(it, maxlen=0)


def resizer(it, size, pad=False, pad_elm=None):
    # type: (Iterable[Sequence[T]], int, bool, Optional[T]) -> Iterable[List[Optional[T]]]

    """Cuts the input iterable `it` consisting of variable length slices
    into slices of length `size`. If pad is True, the last slice
    will be padded to length `size` with `pad_elm`.

    resizer(("asd", "qw", "e"), 4) -> (['a', 's', 'd', 'q'], ['w', 'e'])
    resizer(("asd", "qw", "e"), 4, True, "x") -> (['a', 's', 'd', 'q'], ['w', 'e', 'x', 'x'])
    """

    if size <= 0:
        raise ValueError("size must be larger than 0")

    it = iter(it)
    buf = []  # type: Sequence[T]
    pos = 0
    try:
        while True:
            to_read = size
            out = []  # type: List[Optional[T]]
            while to_read > 0:
                if len(buf) == pos:
                    buf = next(it)
                    pos = 0
                tmp = buf[pos : pos + to_read]
                out.extend(tmp)
                to_read -= len(tmp)
                pos += len(tmp)
            yield out
    except StopIteration:
        if out:
            while pos < len(buf):
                logger.error("this can never happen, can it?")
                tmp = buf[pos : pos + to_read]
                out.extend(tmp)
                to_read -= len(tmp)
                pos += to_read
            if pad:
                yield out + [pad_elm] * (size - len(out))
            else:
                yield out


def switch(it):
    # type: (Iterable[Tuple[T, U]], ) -> Iterator[Tuple[U, T]]

    """Swaps the elements in iterable of pairs `it`.
    ((3, 6), (7, 4), (1, 9)) -> (6, 3), (4, 7), (9, 1)
    """

    return ((j, i) for i, j in it)


def switched_enumerate(it):
    # type: (Iterable[T], ) -> Iterator[Tuple[T, int]]

    """Same as `enumerate()` except that order of the output pair is switched."""

    return switch(enumerate(it))


def multi_join(it_a, it_b, join_func=add):
    # type: (Iterable[T], Iterable[U], Callable[[T, U], V]) -> Iterator[V]

    """Calls `join_func` on the cross product of `it_a` and `it_b`.
    multi_join((1, 2), (3, 4), lambda x, y: str(x)+str(y)) -> ('13', '23', '14', '24')
    """

    return (join_func(a, b) for b in it_b for a in it_a)


def iter_except(iterator, exception_callbacks, return_on_exception=False):
    # type: (Iterator[T], Dict[Type, Callable[[Iterator[T], Exception], bool]], bool) -> Iterator[T]

    """Calls callbacks when exceptions are raised in `iterator`. Does not work for Generators.
    The type key in the `exception_callbacks` dict must be the same type as the exception which
    should be caught, not a subtype.
    """

    if not return_on_exception and isinstance(iterator, GeneratorType):
        raise TypeError("iterator cannot be a generator")

    while True:
        try:
            yield next(iterator)
        except StopIteration:
            return
        except Exception as e:
            try:
                if exception_callbacks[type(e)](
                    iterator, e
                ):  # use if x. and not if not x, so None doesn't raise, if user forgets to returns correct value
                    raise
            except KeyError:
                raise e
            if return_on_exception:
                return


def list_except(it, catch=Exception):
    # type: (Iterable[T], ExceptionsType) -> Tuple[Optional[Exception], List[T]]

    """Same as `list()` except in the case of an exception, the partial list which was collected
    so far is returned. `catch` specifies which exceptions are caught. It can be an exception
    or a tuple of exceptions.
    """

    ret = []  # List[T]
    exc = None  # type: Optional[Exception]

    try:
        for i in it:
            ret.append(i)
    except catch as e:
        exc = e

    return exc, ret


def iter_except_ignore(iterator, exceptions, logger=logger):
    # type: (Iterator[T], ExceptionsType, logging.Logger) -> Iterator[T]

    """Ignores `exceptions` raised in `iterator`. Does not work for Generators."""

    if isinstance(iterator, GeneratorType):
        raise TypeError("iterator cannot be a generator")

    while True:
        try:
            yield next(iterator)
        except StopIteration:
            break
        except exceptions as e:
            logger.warning("Error in iterable: %s", str(e))


def decompress(selectors, data, default=None):
    # type: (Iterable[bool], Iterator[T], Optional[T]) -> Iterator[Optional[T]]

    """Basically the opposite of `itertools.compress`.
    decompress([True, False, True], iter(["A", "B"])) -> ("A", None, "B")
    """

    try:
        for s in selectors:
            if s:
                yield next(data)
            else:
                yield default
    except StopIteration:  # exception converted because StopIterations raised in generators will cause RuntimeErrors
        raise IteratorExhausted


def first_not_none(it, default=None):
    # type: (Iterable[T], Optional[T]) -> Optional[T]

    """Returns the first element of `it` which is not None."""

    return next(filternone(it), default)


def range_count(start=0, stop=None, step=1):
    # type: (int, Optional[int], int) -> Iterable[int]

    """Similar to `range`, except it optionally allows for an infinite range."""

    if stop:
        return range(start, stop, step)
    else:
        return count(start, step)


def product_range_repeat(depth, args):
    # type: (int, Tuple) -> Iterator

    """Returns coordinate or coordinate slices useful for iteration over space.
    Can replace higher dimensional for-in-range loops.
    """

    return product(tuple(range(*args)), repeat=depth)


def powerset(sit):
    # type: (SizedIterable[T], ) -> Iterator[Tuple[T, ...]]

    """Returns the powerset of a sized iterable.
    The powerset of an empty set is a set including the empty set.

    powerset([1,2,3]) --> (), (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    """

    return chain.from_iterable(combinations(sit, r) for r in range(len(sit) + 1))


# had cmp=b"\0" default before
def all_equal_to(it, cmp):
    # type: (Iterable[T], T) -> bool

    """Test if every item of iterable `it` is equal to `cmp`."""

    return all(elm == cmp for elm in it)


def iter_equal(*its):
    # type: (*Iterable) -> bool

    """Test if the contents of all iterables `its` is the same."""

    return all(all_equal(it_elements) for it_elements in zip_longest(*its))


def iter_different(it_a, it_b):
    # type: (Iterable, Iterable) -> bool

    """Test if `it_a` and `it_b` yield different elements."""

    return any(a != b for a, b in zip_longest(it_a, it_b))


def every_n(it, n, pos=0):
    # type: (Iterable[T], int, int) -> Iterator[T]

    """Yields every `n`-th element from iterable `it` and starts at `pos`."""

    if pos > 0:
        advance(it, pos)

    for i, val in enumerate(it):
        if i % n == 0:
            yield val


def split(it, size):
    # type: (Iterable[T], int) -> Sequence[Iterable[T]]

    """memory consumption should be proportional to `size` when output iterators
    are used in parallel loops (like `zip`), else might be higher
    """

    copies = tee(it, size)
    return tuple(every_n(it, size, pos) for it, pos in zip(copies, range(size)))


def no_dupes(*its):
    # type: (*Iterable[T]) -> Iterator[T]

    """Merge `its` and removes all duplicates while preserving order."""

    seen = set()  # type: Set[T]

    for it in its:
        for x in it:
            if x not in seen:
                yield x
                seen.add(x)


def retrier(waittime, attempts=-1, multiplier=1, jitter=0, max_wait=None, jitter_dist="uniform", waitfunc=sleep):
    # type: (float, int, float, float, Optional[float], str, Callable[[float], Any]) -> Iterator[int]

    """Iterator which yields after predefined amounts of time.
    Supports `multiplier` to implements linear or exponential backoff.
    `jitter` adds noise to the average wait time, distributed according to `jitter_dist`.
    Supported values for `jitter_dist` are `uniform` and `normal`.
    """

    if jitter_dist == "uniform":

        def rand(waittime, jitter):
            return waittime + random.uniform(-jitter, jitter)  # nosec

    elif jitter_dist == "normal":

        def rand(waittime, jitter):
            return random.normalvariate(waittime, jitter)

    else:
        raise ValueError("Unsupported jitter_dist")

    if attempts == 0:
        return

    for i in count():
        yield i

        attempts -= 1

        if attempts != 0:
            if max_wait and waittime > max_wait:
                jitter *= max_wait / waittime
                waittime = max_wait
                multiplier = 1

            do_wait = max(0, rand(waittime, jitter))
            waitfunc(do_wait)
        else:
            break

        waittime *= multiplier
        jitter *= multiplier


def collapse_any(it, col_set):
    # type: (Iterable[T], Set[T]) -> Iterator[T]

    """Removes consecutive elements from iterable `it` if element is in `col_set`.
    ([1,1,2,2,3,3,4,4], {1,2}) -> 1, 2, 3, 3, 4, 4
    """

    assert isinstance(col_set, set)

    for key, g in groupby(it):
        if key in col_set:
            yield key
        else:
            yield from g


def collapse_all(it, col_set, replacement):
    # type: (Iterable[T], Set[T], T) -> Iterator[T]

    """Removes consecutive elements from iterable `it` if element is in `col_set`
    and replace `replacement`.
    ([1,1,2,2,3,3,4,4], {1,2}, 5) -> 5, 3, 3, 4, 4
    """

    assert isinstance(col_set, set)

    for key, g in groupby(it, key=operator_in(col_set)):
        if key:
            yield replacement
        else:
            yield from g


def any_in_common(a, b):
    # type: (Iterable[T], Iterable[T]) -> bool

    """Tests if iterables `a` and `b` have any elements in common.
    See also: `any_in`.
    """

    return any(i == j for i, j in product(a, b))


def any_in(it, container):
    # type: (Iterable[T], Container[T]) -> bool

    """Tests if any elements of `it` are in `container`.
    For best performance, `container` should be a set.
    See also: `any_in_common`.
    """

    if isinstance(it, set) and isinstance(container, set):
        return not it.isdisjoint(container)

    return any(elm in container for elm in it)


class CountIter:

    """Example:
    c = CountIter()
    with open("asd.txt", "rt", encoding="utf-8") as fr:
            for line in c.count(fr):
                    func(line)  # do something with line
    print("Processed", c.get(), "lines")
    """

    def __init__(self):
        self.i = 0

    def count(self, it):
        for elm in it:
            self.i += 1
            yield elm

    def get(self):
        return self.i


def select_by_indices(it, indices):
    # type: (Iterable[T], Iterator[int]) -> Iterator[T]

    try:
        sel_ind = next(indices)
        for i, elem in enumerate(it):
            if i == sel_ind:
                yield elem
                sel_ind = next(indices)
    except StopIteration:
        pass


def skip_by_indices(it, indices):
    # type: (Iterator[T], Iterator[int]) -> Iterator[Tuple[int, T]]

    i = 0
    try:
        sel_ind = next(indices)
        for elem in it:
            if i == sel_ind:
                sel_ind = next(indices)
                continue

            yield i, elem
            i += 1
    except StopIteration:
        pass

    for ii, elem in enumerate(it, i):
        yield ii, elem


def remove_consecutive_dupes(it):
    # type: (Iterable, ) -> Iterator

    """Removes consecutive duplicates from iterable `it`.
    (a, a, b, b, a, b) -> a, b, a, b
    """

    return map(itemgetter(0), groupby(it))


def nonstriter(iterable):
    # type: (Iterable, ) -> Iterator

    if isinstance(iterable, (str, bytes)):
        raise TypeError("str or bytes")
    else:
        return iter(iterable)


class CachedIterable(Generic[T]):
    # https://stackoverflow.com/a/19504173

    def __init__(self, iterable):
        # type: (Iterable[T], ) -> None

        self.iterable = iterable
        self.done = False
        self.vals = []  # type: List[T]

    def __iter__(self):
        # type: () -> Iterator[T]

        if self.done:
            return iter(self.vals)
        return chain(self.vals, self._gen_iter())

    def _gen_iter(self):
        # type: () -> Iterator[T]

        for val in self.iterable:
            self.vals.append(val)
            yield val
        self.done = True



from __future__ import generator_stop

from itertools import islice
from typing import TYPE_CHECKING

from nltk.tokenize import word_tokenize as tokenize

from .file import PathOrTextIO

if TYPE_CHECKING:
    from typing import Dict, Iterable, Iterator, List, Optional, TextIO, Union

    from .gensim import KeyedVectors


def gensim_indexer(embeddings, doc, ignore=True):
    # type: (KeyedVectors, str, bool) -> Iterator[int]

    for word in tokenize(doc):
        try:
            yield embeddings.vocab[word.lower()].index
        except KeyError:
            if ignore:
                pass
            else:
                raise


def batch_gensim_indexer(embeddings, docs, ignore=True):
    # type: (KeyedVectors, Iterable[str], bool) -> Iterator[List[int]]

    for doc in docs:
        yield list(gensim_indexer(embeddings, doc, ignore))


def load_freqs(fname, normalize=False, limit=None):
    # type: (Union[str, TextIO], bool, Optional[int]) -> Dict[str, int]

    with PathOrTextIO(fname, "rt", encoding="utf-8") as fin:
        freqs = dict()

        for line in islice(fin, limit):
            word, count = line.split()
            freqs[word] = int(count)

    if normalize:
        total = sum(freqs.keys())

        for word, count in freqs.items():
            freqs[word] = count / total

    return freqs



from __future__ import generator_stop

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List

re_paragraphs = re.compile(r"((?:[^\n][\n]?)+)")


def split_paragraphs(text):
    # type: (str, ) -> List[str]

    return re_paragraphs.findall(text)



from __future__ import generator_stop

from hashlib import sha1
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Union

import bencode

from .exceptions import assert_choice
from .file import blockfileiter, blockfilesiter
from .filesystem import FileProperties

if TYPE_CHECKING:
    from pathlib import Path

    PathType = Union[str, Path]


def read_torrent(path):
    # type: (str, ) -> dict

    return bencode.bread(path)


def write_torrent(data, path):
    # type: (dict, PathType) -> None

    bencode.bwrite(data, path)


def read_torrent_info_dict(path):
    # type: (PathType, ) -> dict

    return bencode.bread(path)["info"]


def iter_torrent(path):
    # type: (PathType, ) -> Iterator[FileProperties]

    """Returns file informations from a torrent file.
    Hash (SHA-1) is obtained according to BEP0047 (if available).
    """

    info = read_torrent_info_dict(path)

    try:
        files = info["files"]
    except KeyError:
        yield FileProperties(info["name"], info["length"], False, hash=info.get("sha1"))
    else:
        for fd in files:
            path = "/".join(fd["path"])
            yield FileProperties(info["name"] + "/" + path, fd["length"], False, hash=fd.get("sha1"))


def pieces_field(pieces):
    return b"".join(sha1(piece).digest() for piece in pieces)  # nosec


def create_torrent_info_dict(path, piece_size, private=None):
    # type: (Path, int, Optional[int]) -> Dict[str, Any]

    if private is not None:
        assert_choice("private", private, {0, 1})

    if path.is_file():
        ret = {
            "name": path.name,
            "length": path.stat().st_size,
            "piece length": piece_size,
            "pieces": pieces_field(blockfileiter(path, chunk_size=piece_size)),
        }

    elif path.is_dir():
        files = []  # fixme: not implemented yet

        assert files, "not implemented yet"

        ret = {
            "name": path.name,
            "files": [
                {
                    "length": length,
                    "path": path.parts,
                }
                for length, path in files
            ],
            "piece length": piece_size,
            "pieces": pieces_field(blockfilesiter((p for _, p in files), chunk_size=piece_size)),
        }

    else:
        raise ValueError("path neither file nor directory")

    if private is not None:
        ret["private"] = private

    return ret


def torrent_info_hash(d):
    return sha1(bencode.bencode(d)).hexdigest()  # nosec


def get_torrent_hash(path):
    return torrent_info_hash(read_torrent_info_dict(path))


def create_torrent(path, piece_size, announce=""):
    # type: (Path, int, str) -> bytes

    info = create_torrent_info_dict(path, piece_size)
    torrent = {"info": info, "announce": announce}

    return bencode.bencode(torrent)



from __future__ import generator_stop

import fcntl
import os
import signal
from typing import TYPE_CHECKING

from .os_shared import _usagetuple

if TYPE_CHECKING:
    from typing import IO


def _lock(fp, exclusive=True, block=False):
    # type: (IO, bool, bool) -> None

    if exclusive:
        operation = fcntl.LOCK_EX
    else:
        operation = fcntl.LOCK_SH

    if not block:
        operation |= fcntl.LOCK_NB

    fcntl.flock(fp, operation)


def _unlock(fp):
    # type: (IO, ) -> None
    fcntl.flock(fp, fcntl.LOCK_UN)


def _disk_usage_posix(path):
    # type: (str, ) -> _usagetuple

    st = os.statvfs(path)

    total = st.f_blocks * st.f_frsize
    free = st.f_bavail * st.f_frsize

    return _usagetuple(total, total - free, free)


def _interrupt_posix():
    os.kill(os.getpid(), signal.SIGINT)


def _filemanager_cmd_posix(path):
    # type: (str, ) -> str

    return f'nautilus "{path}"'  # gnome only. xdg-open for the rest?
